@@ -1,6 +1,4 @@
 #!/usr/bin/env python
-#
-
 __author__ = 'rohe0002'
 
 import requests
@@ -8,12 +6,14 @@ import random
 import string
 import copy
 import cookielib
-from Cookie import SimpleCookie
 import logging
+from Cookie import SimpleCookie
 
 from oic.utils.keyio import KeyJar
 from oic.utils.time_util import utc_time_sans_frac
-from oic.oauth2.exception import UnSupported
+from oic.exception import UnSupported
+
+logger = logging.getLogger(__name__)
 
 DEF_SIGN_ALG = "HS256"
 
@@ -43,8 +43,6 @@ RESPONSE2ERROR = {
 ENDPOINTS = ["authorization_endpoint", "token_endpoint",
              "token_revocation_endpoint"]
 
-logger = logging.getLogger(__name__)
-
 
 class HTTP_ERROR(PyoidcError):
     pass
@@ -147,6 +145,7 @@ class Grant(object):
         self.seed = seed
         self.tokens = []
         self.id_token = None
+        self.code = None
         if resp:
             self.add_code(resp)
             self.add_token(resp)
@@ -210,7 +209,14 @@ class Grant(object):
         return token
 
     def get_id_token(self):
-        return self.id_token
+        if self.id_token:
+            return self.id_token
+        else:
+            for tok in self.tokens:
+                if tok.id_token:
+                    if tok.id_token["exp"] >= time.time():
+                        return tok.id_token
+        return None
 
     def join(self, grant):
         if not self.exp_in:
@@ -272,16 +278,16 @@ def _since_epoch(cdate):
 
 
 class PBase(object):
-    def __init__(self, ca_certs=None):
+    def __init__(self, ca_certs=None, verify_ssl=True):
 
-        self.keyjar = KeyJar()
+        self.keyjar = KeyJar(verify_ssl=verify_ssl)
 
         self.request_args = {"allow_redirects": False}
-        #self.cookies = cookielib.CookieJar()
-        self.cookies = {}
-        self.cookiejar = cookielib.CookieJar()
+        #self.cookies = {}
+        self.cookiejar = cookielib.FileCookieJar()
+        self.ca_certs = ca_certs
         if ca_certs:
-            self.request_args["verify"] = True
+            self.request_args["verify"] = verify_ssl
         else:
             self.request_args["verify"] = False
 
@@ -296,8 +302,9 @@ class PBase(object):
 
         return cookie_dict
 
-    def set_cookie(self, kaka, request):
-        """Returns a cookielib.Cookie based on a set-cookie header line"""
+    def set_cookie(self, kaka):
+        """PLaces a cookie (a cookielib.Cookie based on a set-cookie header
+        line) in the cookie jar. """
 
         # default rfc2109=False
         # max-age, httponly
@@ -366,15 +373,15 @@ class PBase(object):
 
         if self.cookiejar:
             _kwargs["cookies"] = self._cookies()
-            logger.info("SENT COOKIEs: %s" % (_kwargs["cookies"],))
+            logger.debug("SENT COOKIEs: %s" % (_kwargs["cookies"],))
         r = requests.request(method, url, **_kwargs)
         try:
             set_cookie = r.headers["set-cookie"]
             # Telekom fix
             # set_cookie = set_cookie.replace(
             #     "=;Path=/;Expires=Thu, 01-Jan-1970 00:00:01 GMT;HttpOnly,", "")
-            logger.info("RECEIVED COOKIEs: %s" % set_cookie)
-            self.set_cookie(SimpleCookie(set_cookie), r)
+            logger.debug("RECEIVED COOKIEs: %s" % set_cookie)
+            self.set_cookie(SimpleCookie(set_cookie))
         except (AttributeError, KeyError), err:
             pass
 
@@ -383,25 +390,38 @@ class PBase(object):
     def send(self, url, method="GET", **kwargs):
         return self.http_request(url, method, **kwargs)
 
+    def load_cookies_from_file(self, filename, ignore_discard=False,
+                               ignore_expires=False):
+        self.cookiejar.load(filename, ignore_discard, ignore_expires)
+
+    def save_cookies_to_file(self, filename, ignore_discard=False,
+                             ignore_expires=False):
+
+        self.cookiejar.save(filename, ignore_discard, ignore_expires)
+
 
 class Client(PBase):
     _endpoints = ENDPOINTS
 
     def __init__(self, client_id=None, ca_certs=None, client_authn_method=None,
-                 keyjar=None):
+                 keyjar=None, verify_ssl=True):
         """
 
         :param client_id: The client identifier
         :param ca_certs: Certificates used to verify HTTPS certificates
-        :param client_authn_method:
+        :param client_authn_method: Methods that this client can use to
+            authenticate itself. It's a dictionary with method names as
+            keys and method classes as values.
+        :param verify_ssl: Whether the SSL certificate should be verfied.
         :return: Client instance
         """
 
-        PBase.__init__(self, ca_certs)
+        PBase.__init__(self, ca_certs, verify_ssl=verify_ssl)
 
         self.client_id = client_id
         self.client_authn_method = client_authn_method
-        self.keyjar = keyjar or KeyJar()
+        self.keyjar = keyjar or KeyJar(verify_ssl=verify_ssl)
+        self.verify_ssl = verify_ssl
         #self.secret_type = "basic "
 
         self.state = None
@@ -424,6 +444,7 @@ class Client(PBase):
 
         self.provider_info = {}
         self._c_secret = None
+        self.kid = {"sig": {}, "enc": {}}
 
     def get_client_secret(self):
         return self._c_secret
@@ -647,7 +668,7 @@ class Client(PBase):
         return self.construct_request(request, request_args, extra_args)
 
     def get_or_post(self, uri, method, req,
-                    content_type=DEFAULT_POST_CONTENT_TYPE, **kwargs):
+                    content_type=DEFAULT_POST_CONTENT_TYPE, accept=None, **kwargs):
         if method == "GET":
             _qp = req.to_urlencoded()
             if _qp:
@@ -665,7 +686,10 @@ class Client(PBase):
                 raise UnSupported(
                     "Unsupported content type: '%s'" % content_type)
 
-            header_ext = {"content-type": content_type}
+            header_ext = {"Content-type": content_type}
+            if (accept):
+                header_ext = {"Accept": accept}
+            
             if "headers" in kwargs.keys():
                 kwargs["headers"].update(header_ext)
             else:
@@ -714,9 +738,9 @@ class Client(PBase):
 
         if h_arg:
             if "headers" in kwargs.keys():
-                kwargs["headers"].update(h_arg)
+                kwargs["headers"].update(h_arg["headers"])
             else:
-                kwargs["headers"] = h_arg
+                kwargs["headers"] = h_arg["headers"]
 
         return self.uri_and_body(request, cis, method, request_args,
                                  **kwargs)
@@ -826,6 +850,43 @@ class Client(PBase):
         else:
             return http_args
 
+    def parse_request_response(self, reqresp, response, body_type, state="",
+                               **kwargs):
+
+        if reqresp.status_code in [200, 201]:
+            logger.debug("resp.headers: %s" % (reqresp.headers,))
+            logger.debug("resp.txt: %s" % (reqresp.text,))
+            if body_type == "":
+                pass
+            elif body_type == "json":
+                assert "application/json" in reqresp.headers["content-type"]
+            elif body_type == "urlencoded":
+                try:
+                    assert DEFAULT_POST_CONTENT_TYPE in reqresp.headers[
+                        "content-type"]
+                except AssertionError:
+                    assert "text/plain" in reqresp.headers["content-type"]
+            else:
+                raise ValueError("Unknown return format: %s" % body_type)
+        elif reqresp.status_code == 302:  # redirect
+            pass
+        elif reqresp.status_code == 500:
+            logger.error("(%d) %s" % (reqresp.status_code, reqresp.text))
+            raise Exception("ERROR: Something went wrong: %s" % reqresp.text)
+        else:
+            logger.error("(%d) %s" % (reqresp.status_code, reqresp.text))
+            raise Exception("ERROR: Something went wrong: %s [%s]" % (
+                reqresp.text, reqresp.status_code))
+
+        if body_type:
+            if response:
+                return self.parse_response(response, reqresp.text, body_type,
+                                           state, **kwargs)
+            else:
+                raise Exception("Didn't expect a response body")
+        else:
+            return reqresp
+
     def request_and_return(self, url, response=None, method="GET", body=None,
                            body_type="json", state="", http_args=None,
                            **kwargs):
@@ -848,37 +909,11 @@ class Client(PBase):
         except Exception:
             raise
 
-        if resp.status_code in [200, 201]:
-            logger.debug("resp.headers: %s" % (resp.headers,))
-            logger.debug("resp.txt: %s" % (resp.text,))
-            if body_type == "":
-                pass
-            elif body_type == "json":
-                assert "application/json" in resp.headers["content-type"]
-            elif body_type == "urlencoded":
-                try:
-                    assert DEFAULT_POST_CONTENT_TYPE in resp.headers[
-                        "content-type"]
-                except AssertionError:
-                    assert "text/plain" in resp.headers["content-type"]
-            else:
-                raise ValueError("Unknown return format: %s" % body_type)
-        elif resp.status_code == 302:  # redirect
-            pass
-        elif resp.status_code == 500:
-            raise Exception("ERROR: Something went wrong: %s" % resp.text)
-        else:
-            raise Exception("ERROR: Something went wrong: %s [%s]" % (
-                resp.text, resp.status_code))
+        if not "keyjar" in kwargs:
+            kwargs["keyjar"] = self.keyjar
 
-        if body_type:
-            if response:
-                return self.parse_response(response, resp.text, body_type,
-                                           state, **kwargs)
-            else:
-                raise Exception("Didn't expect a response body")
-        else:
-            return resp
+        return self.parse_request_response(resp, response, body_type, state,
+                                           **kwargs)
 
     def do_authorization_request(self, request=AuthorizationRequest,
                                  state="", body_type="", method="GET",
@@ -1029,10 +1064,11 @@ class Client(PBase):
 
 
 class Server(PBase):
-    def __init__(self, keys=None, ca_certs=None):
-        PBase.__init__(self, ca_certs)
+    def __init__(self, keys=None, ca_certs=None, verify_ssl=True):
+        PBase.__init__(self, ca_certs, verify_ssl)
 
-    def parse_url_request(self, request, url=None, query=None):
+    @staticmethod
+    def parse_url_request(request, url=None, query=None):
         if url:
             parts = urlparse.urlparse(url)
             scheme, netloc, path, params, query, fragment = parts[:6]
