@@ -8,22 +8,19 @@ import jwkest
 from jwkest.jwe import JWE
 from jwkest.jwk import keyitems2keyreps
 from jwkest.jws import JWS
-from oic.oauth2.exception import PyoidcError
+from oic.exception import PyoidcError
+from oic.exception import MessageException
 
 logger = logging.getLogger(__name__)
 
 
-class MessageException(PyoidcError):
-    pass
-
-
 class MissingRequiredAttribute(MessageException):
-    def __init__(self, attr):
-        Exception.__init__(self)
-        self.attr = attr
+    def __init__(self, attr, message=""):
+        Exception.__init__(self, attr)
+        self.message = message
 
     def __str__(self):
-        return "Missing required attribute '%s'" % self.attr
+        return "Missing required attribute '%s'" % self.args[0]
 
 
 class MissingRequiredValue(MessageException):
@@ -46,6 +43,8 @@ class OldAccessToken(PyoidcError):
     pass
 
 
+class SchemeError(MessageException):
+    pass
 
 
 ERRTXT = "On '%s': %s"
@@ -116,7 +115,7 @@ class Message(object):
         if not self.lax:
             for attribute, (_, req, _ser, _, na) in _spec.items():
                 if req and attribute not in self._dict:
-                    raise MissingRequiredAttribute("%s" % attribute)
+                    raise MissingRequiredAttribute("%s" % attribute, self)
 
         params = []
 
@@ -173,9 +172,11 @@ class Message(object):
 
     def deserialize(self, info, method="urlencoded", **kwargs):
         try:
-            return getattr(self, "from_%s" % method)(info, **kwargs)
+            func = getattr(self, "from_%s" % method)
         except AttributeError, err:
             raise Exception("Unknown method (%s)" % method)
+        else:
+            return func(info, **kwargs)
 
     def from_urlencoded(self, urlencoded, **kwargs):
         """
@@ -229,7 +230,7 @@ class Message(object):
                         try:
                             self._dict[key] = typ(val[0])
                         except KeyError:
-                            raise ValueError
+                            raise ValueError()
                 else:
                     raise TooManyValues
 
@@ -452,7 +453,7 @@ class Message(object):
                 dkeys = keyjar.get_decrypt_key(owner="")
             else:
                 dkeys = {}
-            txt = JWE().decrypt(txt, dkeys, "private")
+            txt = JWE().decrypt(txt, dkeys)
             try:
                 jso = json.loads(txt)
             except Exception:
@@ -476,9 +477,12 @@ class Message(object):
                             pass
 
                 if _kid:
-                    _key = keyjar.get_key_by_kid(_kid, jso["iss"])
-                    if _key:
-                        key.append(_key)
+                    try:
+                        _key = keyjar.get_key_by_kid(_kid, jso["iss"])
+                        if _key:
+                            key.append(_key)
+                    except KeyError:
+                        pass
 
                 try:
                     self._add_key(keyjar, kwargs["opponent_id"], key)
@@ -586,11 +590,14 @@ class Message(object):
     def __contains__(self, item):
         return item in self._dict
 
-    def request(self, location):
-        if "?" in location:
-            return "%s&%s" % (location, self.to_urlencoded())
+    def request(self, location, fragment_enc=False):
+        if fragment_enc:
+            return "%s#%s" % (location, self.to_urlencoded())
         else:
-            return "%s?%s" % (location, self.to_urlencoded())
+            if "?" in location:
+                return "%s&%s" % (location, self.to_urlencoded())
+            else:
+                return "%s?%s" % (location, self.to_urlencoded())
 
     def __setitem__(self, key, value):
         try:
@@ -652,7 +659,8 @@ class Message(object):
     def from_jwe(self, msg, keys):
         krs = keyitems2keyreps(keys)
         jwe = JWE()
-        return self.from_json(jwe.decrypt(msg, krs))
+        _res = jwe.decrypt(msg, krs)
+        return self.from_json(_res[0])
 
 # =============================================================================
 
@@ -671,6 +679,8 @@ def add_non_standard(msg1, msg2):
 
 #noinspection PyUnusedLocal
 def list_serializer(vals, sformat="urlencoded", lev=0):
+    if isinstance(vals, basestring) or not isinstance(vals, list):
+        raise ValueError("Expected list: %s" % vals)
     if sformat == "urlencoded":
         return " ".join(vals)
     else:
