@@ -3,6 +3,7 @@ import traceback
 import sys
 import urllib
 import urlparse
+from oic.utils.sdb import AccessCodeUsed
 
 __author__ = 'rohe0002'
 
@@ -10,11 +11,12 @@ import base64
 import logging
 import os
 
-from oic.oauth2.exception import MissingParameter, URIError
-from oic.oauth2.exception import RedirectURIError
-from oic.oauth2.exception import ParameterError
-from oic.oauth2.exception import FailedAuthentication
-from oic.oauth2.exception import UnknownClient
+from oic.exception import MissingParameter
+from oic.exception import URIError
+from oic.exception import RedirectURIError
+from oic.exception import ParameterError
+from oic.exception import FailedAuthentication
+from oic.exception import UnknownClient
 
 from oic.oauth2.message import AccessTokenResponse
 from oic.oauth2.message import ErrorResponse
@@ -28,7 +30,8 @@ from oic.oauth2.message import MissingRequiredAttribute
 from oic.oauth2.message import TokenErrorResponse
 from oic.oauth2.message import AccessTokenRequest
 
-from oic.utils.http_util import BadRequest, CookieDealer
+from oic.utils.http_util import BadRequest
+from oic.utils.http_util import CookieDealer
 from oic.utils.http_util import make_cookie
 from oic.utils.http_util import Redirect
 from oic.utils.http_util import Response
@@ -82,7 +85,7 @@ def token_response(**kwargs):
     _areq = kwargs["areq"]
     _scode = kwargs["scode"]
     _sdb = kwargs["sdb"]
-    _dic = _sdb.update_to_token(_scode, issue_refresh=False)
+    _dic = _sdb.upgrade_to_token(_scode, issue_refresh=False)
 
     aresp = AccessTokenResponse(**by_schema(AccessTokenResponse, **_dic))
     if "state" in _areq:
@@ -113,11 +116,11 @@ class Provider(object):
 
     def __init__(self, name, sdb, cdb, authn_broker, authz, client_authn,
                  symkey="", urlmap=None, iv=0, default_scope="",
-                 ca_bundle=None):
+                 ca_bundle=None, verify_ssl=True, default_acr=""):
         self.name = name
         self.sdb = sdb
         self.cdb = cdb
-        self.server = Server(ca_certs=ca_bundle)
+        self.server = Server(ca_certs=ca_bundle, verify_ssl=verify_ssl)
 
         self.authn_broker = authn_broker
         if authn_broker is None:
@@ -136,6 +139,7 @@ class Provider(object):
         self.cookie_name = "pyoidc"
         self.default_scope = default_scope
         self.sso_ttl = 0
+        self.default_acr = default_acr
 
         if urlmap is None:
             self.urlmap = {}
@@ -217,10 +221,10 @@ class Provider(object):
     def authn_reply(self, areq, aresp, bsid, **kwargs):
         """
 
-        :param areq:
-        :param aresp:
-        :param bsid:
-        :param kwargs:
+        :param areq: Authorization Request
+        :param aresp: Authorization Response
+        :param bsid: Session id
+        :param kwargs: Additional keyword args
         :return:
         """
         if "redirect_uri" in areq:
@@ -253,7 +257,8 @@ class Provider(object):
         return self.response_type_map[_rtype](areq=areq, scode=scode,
                                               sdb=self.sdb)
 
-    def input(self, query="", post=None):
+    @staticmethod
+    def input(query="", post=None):
         # Support GET and POST
         if query:
             return query
@@ -262,18 +267,21 @@ class Provider(object):
         else:
             raise MissingParameter("No input")
 
-    def _error_response(self, error, descr=None):
+    @staticmethod
+    def _error_response(error, descr=None):
         logger.error("%s" % error)
         response = ErrorResponse(error=error, error_description=descr)
         return Response(response.to_json(), content="application/json",
                         status="400 Bad Request")
 
-    def _error(self, error, descr=None):
+    @staticmethod
+    def _error(error, descr=None):
         response = ErrorResponse(error=error, error_description=descr)
         return Response(response.to_json(), content="application/json",
                         status="400 Bad Request")
 
-    def _authz_error(self, error, descr=None):
+    @staticmethod
+    def _authz_error(error, descr=None):
 
         response = AuthorizationErrorResponse(error=error)
         if descr:
@@ -282,7 +290,8 @@ class Provider(object):
         return Response(response.to_json(), content="application/json",
                         status="400 Bad Request")
 
-    def _redirect_authz_error(self, error, redirect_uri, descr=None):
+    @staticmethod
+    def _redirect_authz_error(error, redirect_uri, descr=None):
         err = ErrorResponse(error=error)
         if descr:
             err["error_description"] = descr
@@ -351,15 +360,43 @@ class Provider(object):
 
         return uri
 
-    def pick_auth(self, **kwargs):
+    def pick_auth(self, areq, comparision_type=""):
         """
-        Pick authentication method to be used
-        :param kwargs: Extra key word arguments
-        :return: Authentication method
-        """
-        return self.authn_broker[0]
 
-    def authorization_endpoint(self, request="", cookie=None, **kwargs):
+        :param areq: AuthorizationRequest instance
+        :param comparision_type: How to pick the authentication method
+        :return: An authentication method and its authn class ref
+        """
+        if comparision_type == "any":
+            return self.authn_broker[0]
+
+        try:
+            if len(self.authn_broker) == 1:
+                    return self.authn_broker[0]
+            else:
+                try:
+                    _values = areq["acr_values"]
+                except KeyError:
+                    _values = self.default_acr
+
+                if isinstance(_values, basestring):
+                    _values = [_values]
+
+                if not comparision_type:
+                    comparision_type = "exact"
+
+                for _acr in _values:
+                    res = self.authn_broker.pick(_acr, comparision_type)
+                    if res:
+                        #Return the best guess by pick.
+                        return res[0]
+        except KeyError:
+            pass
+
+        # return the best I have
+        return None, None
+
+    def authorization_endpoint(self, request="", cookie="", authn="", **kwargs):
         """ The AuthorizationRequest endpoint
 
         :param request: The client request
@@ -392,8 +429,10 @@ class Provider(object):
         logger.debug("AuthzRequest: %s" % (areq.to_dict(),))
         try:
             redirect_uri = self.get_redirect_uri(areq)
-        except (RedirectURIError, ParameterError, UnknownClient), err:
+        except (RedirectURIError, ParameterError), err:
             return self._error("invalid_request", "%s" % err)
+        except UnknownClient, err:
+            return self._error("unauthorized_client", "%s" % err)
 
         try:
             # verify that the request message is correct
@@ -407,15 +446,18 @@ class Provider(object):
         if cookie:
             logger.debug("Cookie: %s" % cookie)
             a_args["cookie"] = cookie
-        if "authn" in kwargs:
+        if authn:
             try:
-                a_args["authorization"] = kwargs["authn"]
+                a_args["authorization"] = authn
             except KeyError:
                 pass
 
         # Pick authentication method
         _authn, acr = self.pick_auth(areq=areq)
 
+        if authn is None:
+            return self._redirect_authz_error("invalid_request", redirect_uri)
+
         try:
             identity = _authn.authenticated_as(**a_args)
         except (NoSuchAuthentication, ToOld, TamperAllert):
@@ -451,26 +493,26 @@ class Provider(object):
         except KeyError:
             oidc_req = None
 
-        sinfo = self.sdb.create_authz_session(user, areq, oidreq=oidc_req)
+        skey = self.sdb.create_authz_session(user, areq, oidreq=oidc_req)
 
         # Now about the authorization step.
         try:
             permissions = self.authz.permissions(cookie)
             if not permissions:
-                return self.authz(user, sinfo)
+                return self.authz(user, skey)
         except (ToOld, TamperAllert):
-            return self.authz(user, areq, sinfo)
+            return self.authz(user, areq, skey)
 
-        return self.authz_part2(user, areq, sinfo, permissions, _authn)
+        return self.authz_part2(user, areq, skey, permissions, _authn)
 
-    def authz_part2(self, user, areq, sinfo, permission=None, authn=None,
+    def authz_part2(self, user, areq, skey, permission=None, authn=None,
                     **kwargs):
         """
         After the authentication this is where you should end up
 
         :param user:
         :param areq: The Authorization Request
-        :param sinfo: Session information
+        :param skey: Session key
         :param permission: A permission specification
         :param authn: The Authentication Method used
         :param kwargs: possible other parameters
@@ -479,7 +521,7 @@ class Provider(object):
         _log_debug = logger.debug
         _log_debug("- in authenticated() -")
 
-        sinfo["auz"] = permission
+        self.sdb.update(skey, "auz", permission)
 
         _log_debug("response type: %s" % areq["response_type"])
 
@@ -504,7 +546,7 @@ class Provider(object):
             except KeyError:
                 pass
 
-            _log_debug("_dic: %s" % sinfo)
+            _log_debug("_dic: %s" % self.sdb[skey])
 
             rtype = set(areq["response_type"][:])
             if "code" in areq["response_type"]:
@@ -512,19 +554,20 @@ class Provider(object):
                 #    scode = self.sdb.duplicate(_sinfo)
                 #    _sinfo = self.sdb[scode]
 
-                _code = aresp["code"] = self.sdb.get_token(sinfo)
+                _code = aresp["code"] = self.sdb.get_token(skey)
                 rtype.remove("code")
             else:
-                sinfo["code"] = None
-                _code = None
+                _code = self.sdb[skey]["code"]
+                self.sdb.update(skey, "code", None)
 
             if "token" in rtype:
-                _dic = self.sdb.upgrade_to_token(sinfo, issue_refresh=False)
-
+                self.sdb.upgrade_to_token(skey, issue_refresh=False,
+                                          access_grant=_code)
                 atr = AccessTokenResponse(**aresp.to_dict())
                 aresp = atr
-                _log_debug("_dic: %s" % _dic)
-                for key, val in _dic.items():
+                _cont = self.sdb[skey]
+                _log_debug("_dic: %s" % _cont)
+                for key, val in _cont.items():
                     if key in aresp.parameters() and val is not None:
                         aresp[key] = val
 
@@ -538,7 +581,7 @@ class Provider(object):
         except (RedirectURIError, ParameterError), err:
             return BadRequest("%s" % err)
 
-        self.sdb.store_session(sinfo)
+        #self.sdb.store_session(skey)
 
         # so everything went well should set a SSO cookie
         headers = [authn.create_cookie(user, typ="sso", ttl=self.sso_ttl)]
@@ -554,7 +597,7 @@ class Provider(object):
         #     return Response(err.to_json(), content="application/json")
         return None
 
-    def token_endpoint(self, auth_header="", **kwargs):
+    def token_endpoint(self, authn="", **kwargs):
         """
         This is where clients come to get their access tokens
         """
@@ -568,7 +611,7 @@ class Provider(object):
         areq = AccessTokenRequest().deserialize(body, "urlencoded")
 
         try:
-            client = self.client_authn(self, areq, auth_header)
+            client = self.client_authn(self, areq, authn)
         except FailedAuthentication, err:
             err = TokenErrorResponse(error="unathorized_client",
                                      error_description="%s" % err)
@@ -577,7 +620,14 @@ class Provider(object):
 
         LOG_DEBUG("AccessTokenRequest: %s" % areq)
 
-        assert areq["grant_type"] == "authorization_code"
+        try:
+            assert areq["grant_type"] == "authorization_code"
+        except AssertionError:
+            err = TokenErrorResponse(error="invalid_request",
+                                     error_description="Wrong grant type")
+            return Response(err.to_json(), content="application/json",
+                            status="401 Unauthorized")
+
 
         # assert that the code is valid
         _info = _sdb[areq["code"]]
@@ -591,7 +641,14 @@ class Provider(object):
         if "redirect_uri" in _info:
             assert areq["redirect_uri"] == _info["redirect_uri"]
 
-        _tinfo = _sdb.update_to_token(areq["code"])
+        try:
+            _tinfo = _sdb.upgrade_to_token(areq["code"])
+        except AccessCodeUsed:
+            err = TokenErrorResponse(error="invalid_grant",
+                                     error_description="Access grant used")
+            return Response(err.to_json(), content="application/json",
+                            status="401 Unauthorized")
+
 
         LOG_DEBUG("_tinfo: %s" % _tinfo)
             
