@@ -39,7 +39,10 @@ from oic.oauth2 import HTTP_ARGS
 from oic.oauth2 import rndstr
 from oic.oauth2.consumer import ConfigurationError
 
-from oic.oauth2.exception import AccessDenied, PyoidcError, MissingParameter
+from oic.exception import AccessDenied
+from oic.exception import IssuerMismatch
+from oic.exception import PyoidcError
+from oic.exception import MissingParameter
 
 from oic.utils import time_util
 
@@ -82,17 +85,19 @@ REQUEST2ENDPOINT = {
 }
 
 # -----------------------------------------------------------------------------
-MAX_AUTHENTICATION_AGE = 86400
+
 JWT_BEARER = "urn:ietf:params:oauth:client-assertion-type:jwt-bearer"
+SAML2_BEARER_GRANT_TYPE = "urn:ietf:params:oauth:grant-type:saml2-bearer"
 
 OIDCONF_PATTERN = "%s/.well-known/openid-configuration"
 
+# This should probably be part of the configuration
+MAX_AUTHENTICATION_AGE = 86400
 DEF_SIGN_ALG = {"id_token": "RS256",
                 "openid_request_object": "RS256",
                 "client_secret_jwt": "HS256",
                 "private_key_jwt": "RS256"}
 
-SAML2_BEARER_GRANT_TYPE = "urn:ietf:params:oauth:grant-type:saml2-bearer"
 
 
 # -----------------------------------------------------------------------------
@@ -115,7 +120,7 @@ def verify_acr_level(req, level):
     else:  # Required or Optional
         return level
 
-    raise AccessDenied
+    raise AccessDenied("", req)
 
 
 def deser_id_token(inst, txt=""):
@@ -189,7 +194,7 @@ class Grant(oauth2.Grant):
 
 
 PREFERENCE2PROVIDER = {
-    "require_signed_request_object": "request_object_algs_supported",
+    #"require_signed_request_object": "request_object_algs_supported",
     "request_object_signing_alg": "request_object_signing_alg_values_supported",
     "request_object_encryption_alg":
         "request_object_encryption_alg_values_supported",
@@ -210,7 +215,8 @@ PREFERENCE2PROVIDER = {
     "subject_type": "subject_types_supported",
     "token_endpoint_auth_method": "token_endpoint_auth_methods_supported",
     "token_endpoint_auth_signing_alg":
-        "token_endpoint_auth_signing_alg_values_supported"
+        "token_endpoint_auth_signing_alg_values_supported",
+    "response_types": "response_types_supported"
     #"request_object_signing_alg": "request_object_signing_alg_values_supported
 }
 
@@ -225,11 +231,12 @@ class Client(oauth2.Client):
     _endpoints = ENDPOINTS
 
     def __init__(self, client_id=None, ca_certs=None,
-                 client_prefs=None, client_authn_method=None, keyjar=None):
+                 client_prefs=None, client_authn_method=None, keyjar=None,
+                 verify_ssl=True):
 
         oauth2.Client.__init__(self, client_id, ca_certs,
                                client_authn_method=client_authn_method,
-                               keyjar=keyjar)
+                               keyjar=keyjar, verify_ssl=verify_ssl)
 
         self.file_store = "./file/"
         self.file_uri = "http://localhost/"
@@ -243,18 +250,27 @@ class Client(oauth2.Client):
 
         self.request2endpoint = REQUEST2ENDPOINT
         self.response2error = RESPONSE2ERROR
+
         self.grant_class = Grant
         self.token_class = Token
-        self.provider_info = {}
+        self.provider_info = None
+        self.registration_response = None
         self.client_prefs = client_prefs or {}
+
         self.behaviour = {
-            "require_signed_request_object":
+            "request_object_signing_alg":
             DEF_SIGN_ALG["openid_request_object"]}
 
         self.wf = WebFinger(OIC_ISSUER)
         self.wf.httpd = self
         self.allow = {}
         self.post_logout_redirect_uris = []
+        self.registration_expires = 0
+        self.registration_access_token = None
+
+        # Default key by kid for different key types
+        # For instance {"RSA":"abc"}
+        self.kid = {"sig": {}, "enc": {}}
 
     def _get_id_token(self, **kwargs):
         try:
@@ -313,13 +329,17 @@ class Client(oauth2.Client):
                                                             **kwargs)
 
         if request_param:
-            alg = self.behaviour["require_signed_request_object"]
+            alg = self.behaviour["request_object_signing_alg"]
             if "algorithm" not in kwargs:
                 kwargs["algorithm"] = alg
 
             if "keys" not in kwargs and alg:
-                atype = alg2keytype(alg)
-                kwargs["keys"] = self.keyjar.get_signing_key(atype)
+                _kty = alg2keytype(alg)
+                try:
+                    kwargs["keys"] = self.keyjar.get_signing_key(
+                        _kty, kid=self.kid["sig"][_kty])
+                except KeyError:
+                    kwargs["keys"] = self.keyjar.get_signing_key(_kty)
 
             _req = make_openid_request(areq, **kwargs)
 
@@ -586,7 +606,7 @@ class Client(oauth2.Client):
             if kwargs["token"]:
                 uir["access_token"] = kwargs["token"]
                 token = Token()
-                token.type = "Bearer"
+                token.token_type = "Bearer"
                 token.access_token = kwargs["token"]
                 kwargs["behavior"] = "use_authorization_header"
             else:
@@ -601,6 +621,8 @@ class Client(oauth2.Client):
 
             if token.is_valid():
                 uir["access_token"] = token.access_token
+                if token.token_type == "Bearer" and method == "GET":
+                    kwargs["behavior"] = "use_authorization_header"
             else:
                 # raise oauth2.OldAccessToken
                 if self.log:
@@ -626,7 +648,7 @@ class Client(oauth2.Client):
                 _ttype = kwargs["token_type"]
             except KeyError:
                 try:
-                    _ttype = token.type
+                    _ttype = token.token_type
                 except AttributeError:
                     raise MissingParameter("Unspecified token type")
 
@@ -687,8 +709,13 @@ class Client(oauth2.Client):
             return _schema().from_json(txt=resp.text)
         else:
             algo = self.client_prefs["userinfo_signed_response_alg"]
+            _kty = alg2keytype(algo)
             # Keys of the OP ?
-            keys = self.keyjar.get_signing_key(alg2keytype(algo))
+            try:
+                keys = self.keyjar.get_signing_key(_kty, self.kid["sig"][_kty])
+            except KeyError:
+                keys = self.keyjar.get_signing_key(_kty)
+
             return _schema().from_jwt(resp.text, keys)
 
     def get_userinfo_claims(self, access_token, endpoint, method="POST",
@@ -730,7 +757,7 @@ class Client(oauth2.Client):
         :param issuer: The one I thought should be the issuer of the config
         :param keys: Should I deal with keys
         :param endpoints: Should I deal with endpoints, that is store them
-            as attributes in self.
+        as attributes in self.
         """
 
         if "issuer" in pcr:
@@ -752,11 +779,11 @@ class Client(oauth2.Client):
                 try:
                     assert _issuer == _pcr_issuer
                 except AssertionError:
-                    raise PyoidcError(
-                        "provider info issuer mismatch '%s' != '%s'" % (
-                            _issuer, _pcr_issuer))
+                    raise IssuerMismatch("'%s' != '%s'" % (_issuer,
+                                                           _pcr_issuer),
+                                         pcr)
 
-            self.provider_info[_pcr_issuer] = pcr
+            self.provider_info = pcr
         else:
             _pcr_issuer = issuer
 
@@ -767,7 +794,7 @@ class Client(oauth2.Client):
 
         if keys:
             if self.keyjar is None:
-                self.keyjar = KeyJar()
+                self.keyjar = KeyJar(verify_ssl=self.verify_ssl)
 
             self.keyjar.load_keys(pcr, _pcr_issuer)
 
@@ -792,6 +819,7 @@ class Client(oauth2.Client):
                     pcr = response_cls().from_json(r.text)
                     break
 
+        #logger.debug("Provider info: %s" % pcr)
         if pcr is None:
             raise PyoidcError("Trying '%s', status %s" % (url, r.status_code))
 
@@ -862,7 +890,7 @@ class Client(oauth2.Client):
         """
 
         try:
-            _pcr = self.provider_info[other]
+            _pcr = self.provider_info
             supported = _pcr["%s_algs_supported" % usage]
         except KeyError:
             try:
@@ -886,7 +914,9 @@ class Client(oauth2.Client):
         :param issuer: The issuer identifier
         """
         if not pcr:
-            pcr = self.provider_info[issuer]
+            pcr = self.provider_info
+
+        regreq = RegistrationRequest
 
         for _pref, _prov in PREFERENCE2PROVIDER.items():
             try:
@@ -901,26 +931,44 @@ class Client(oauth2.Client):
                     self.behaviour[_pref] = PROVIDER_DEFAULT[_pref]
                 except KeyError:
                     #self.behaviour[_pref]= vals[0]
-                    self.behaviour[_pref] = None
+                    if isinstance(pcr.c_param[_prov][0], list):
+                        self.behaviour[_pref] = []
+                    else:
+                        self.behaviour[_pref] = None
                 continue
 
-            for val in vals:
-                if val in _pvals:
-                    self.behaviour[_pref] = val
-                    break
+            if isinstance(vals, basestring):
+                if vals in _pvals:
+                    self.behaviour[_pref] = vals
+            else:
+                vtyp = regreq.c_param[_pref]
+                if isinstance(vtyp[0], list):
+                    _list = True
+                else:
+                    _list = False
+
+                for val in vals:
+                    if val in _pvals:
+                        if not _list:
+                            self.behaviour[_pref] = val
+                            break
+                        else:
+                            try:
+                                self.behaviour[_pref].append(val)
+                            except KeyError:
+                                self.behaviour[_pref] = [val]
 
             if _pref not in self.behaviour:
                 raise ConfigurationError(
-                    "OP couldn't match preferences", "%s" % _pref)
+                    "OP couldn't match preference:%s" % _pref, pcr)
 
-        regreq = RegistrationRequest
         for key, val in self.client_prefs.items():
             if key in self.behaviour:
                 continue
 
             try:
                 vtyp = regreq.c_param[key]
-                if isinstance(vtyp, list):
+                if isinstance(vtyp[0], list):
                     pass
                 elif isinstance(val, list) and not isinstance(val, basestring):
                     val = val[0]
@@ -931,10 +979,19 @@ class Client(oauth2.Client):
 
     def store_registration_info(self, reginfo):
         self.registration_response = reginfo
+        if "token_endpoint_auth_method" not in self.registration_response:
+            self.registration_response["token_endpoint_auth_method"] = "client_secret_post"
         self.client_secret = reginfo["client_secret"]
         self.client_id = reginfo["client_id"]
-        self.registration_expires = reginfo["client_secret_expires_at"]
-        self.registration_access_token = reginfo["registration_access_token"]
+        try:
+            self.registration_expires = reginfo["client_secret_expires_at"]
+        except KeyError:
+            pass
+        try:
+            self.registration_access_token = reginfo[
+                "registration_access_token"]
+        except KeyError:
+            pass
 
     def handle_registration_info(self, response):
         if response.status_code == 200:
@@ -987,7 +1044,7 @@ class Client(oauth2.Client):
             try:
                 req["redirect_uris"] = self.redirect_uris
             except AttributeError:
-                raise MissingRequiredAttribute("redirect_uris")
+                raise MissingRequiredAttribute("redirect_uris", req)
 
         return req
 
@@ -1029,8 +1086,8 @@ class Client(oauth2.Client):
 
 #noinspection PyMethodOverriding
 class Server(oauth2.Server):
-    def __init__(self, keyjar=None, ca_certs=None):
-        oauth2.Server.__init__(self, keyjar, ca_certs)
+    def __init__(self, keyjar=None, ca_certs=None, verify_ssl=True):
+        oauth2.Server.__init__(self, keyjar, ca_certs, verify_ssl)
 
     def _parse_urlencoded(self, url=None, query=None):
         if url:
@@ -1090,13 +1147,17 @@ class Server(oauth2.Server):
                 query = data
             request = request().from_urlencoded(query)
         else:
-            raise PyoidcError("Unknown package format: '%s'" % sformat)
+            raise PyoidcError("Unknown package format: '%s'" % sformat,
+                              request)
 
         # get the verification keys
         if client_id:
             keys = self.keyjar.verify_keys(client_id)
         else:
-            keys = None
+            try:
+                keys = self.keyjar.verify_keys(request["client_id"])
+            except KeyError:
+                keys = None
 
         request.verify(key=keys, keyjar=self.keyjar)
         return request
@@ -1127,24 +1188,38 @@ class Server(oauth2.Server):
     #    def parse_issuer_request(self, info, sformat="urlencoded"):
     #        return self._parse_request(IssuerRequest, info, sformat)
 
-    def id_token_claims(self, session):
+    @staticmethod
+    def id_token_claims(session):
         """
         Pick the IdToken claims from the request
 
         :param session: Session information
         :return: The IdToken claims
         """
+        itc = None
+
+        try:
+            authzreq = AuthorizationRequest().deserialize(session["authzreq"],
+                                                          'json')
+            itc = authzreq["claims"]["id_token"]
+            logger.debug("ID Token claims: %s" % itc)
+        except KeyError:
+            pass
+
         try:
             oidreq = OpenIDRequest().deserialize(session["oidreq"], "json")
-            itc = oidreq["claims"]["id_token"]
+            itc_or = oidreq["claims"]["id_token"]
+            if itc:
+                itc.update(itc_or)
             logger.debug("ID Token claims: %s" % itc)
-            return itc
         except KeyError:
-            return None
+            pass
+
+        return itc
 
     def make_id_token(self, session, loa="2", issuer="",
                       alg="RS256", code=None, access_token=None,
-                      user_info=None):
+                      user_info=None, auth_time=0):
         """
 
         :param session: Session information
@@ -1168,7 +1243,7 @@ class Server(oauth2.Server):
                 inawhile = {}
             for key, val in itc.items():
                 if key == "auth_time":
-                    extra["auth_time"] = time_util.utc_time_sans_frac()
+                    extra["auth_time"] = auth_time
                 elif key == "acr":
                     #["2","http://id.incommon.org/assurance/bronze"]
                     extra["acr"] = verify_acr_level(val, loa)
@@ -1206,6 +1281,6 @@ class Server(oauth2.Server):
             idt[key] = val
 
         if "nonce" in session:
-            idt.nonce = session["nonce"]
+            idt["nonce"] = session["nonce"]
 
         return idt
