@@ -1,6 +1,5 @@
 import time
 from urlparse import urlparse
-from oic.oauth2.exception import InvalidRequest, PyoidcError
 
 __author__ = 'rohe0002'
 
@@ -8,10 +7,15 @@ import urllib
 import json
 import logging
 
-from oic.oauth2 import message, MissingRequiredValue
+from oic.oauth2 import message
+from oic.oauth2 import MissingRequiredValue
 from oic.oauth2 import MissingRequiredAttribute
 from oic.oauth2 import VerificationError
-from oic.oauth2.message import Message, REQUIRED_LIST_OF_SP_SEP_STRINGS
+from oic.exception import InvalidRequest, NotForMe
+from oic.exception import PyoidcError
+from oic.oauth2.message import Message, SchemeError
+from oic.oauth2.message import REQUIRED_LIST_OF_SP_SEP_STRINGS
+from oic.oauth2.message import SINGLE_OPTIONAL_JSON
 from oic.oauth2.message import SINGLE_OPTIONAL_STRING
 from oic.oauth2.message import OPTIONAL_LIST_OF_STRINGS
 from oic.oauth2.message import SINGLE_REQUIRED_STRING
@@ -37,7 +41,6 @@ def json_deser(val, sformat=None, lev=0):
 # value type, required, serializer, deserializer, null value allowed
 SINGLE_OPTIONAL_BOOLEAN = (bool, False, None, None, False)
 SINGLE_OPTIONAL_JSON_WN = (dict, False, json_ser, json_deser, True)
-SINGLE_OPTIONAL_JSON = (dict, False, json_ser, json_deser, False)
 SINGLE_REQUIRED_INT = (int, True, None, None, False)
 
 
@@ -92,7 +95,7 @@ def msg_ser(inst, sformat, lev=0):
         else:
             raise ValueError("%s" % type(inst))
     else:
-        raise PyoidcError("Unknown sformat")
+        raise PyoidcError("Unknown sformat", inst)
 
     return res
 
@@ -105,7 +108,7 @@ def msg_ser_json(inst, sformat="json", lev=0):
         elif isinstance(inst, dict):
             res = inst
         else:
-            raise ValueError("%s" % type(inst))
+            raise ValueError("%s" % type(inst), inst)
     else:
         sformat = "json"
         if isinstance(inst, dict) or isinstance(inst, Message):
@@ -145,7 +148,7 @@ def claims_ser(val, sformat="urlencoded", lev=0):
         else:
             raise ValueError("%s" % type(item))
     else:
-        raise PyoidcError("Unknown sformat")
+        raise PyoidcError("Unknown sformat: %s" % sformat, val)
 
     return res
 
@@ -249,7 +252,7 @@ class AuthorizationResponse(message.AuthorizationResponse,
     c_param.update(message.AccessTokenResponse.c_param)
     c_param.update({
         "code": SINGLE_OPTIONAL_STRING,
-        "nonce": SINGLE_OPTIONAL_STRING,
+        #"nonce": SINGLE_OPTIONAL_STRING,
         "access_token": SINGLE_OPTIONAL_STRING,
         "token_type": SINGLE_OPTIONAL_STRING,
         "id_token": SINGLE_OPTIONAL_IDTOKEN
@@ -272,7 +275,7 @@ class AuthorizationResponse(message.AuthorizationResponse,
                     pass
             idt = IdToken().from_jwt(str(self["id_token"]), **args)
             if not idt.verify(**kwargs):
-                raise VerificationError("Could not verify id_token")
+                raise VerificationError("Could not verify id_token", idt)
 
             hfunc = "HS" + jwkest.unpack(self["id_token"])[0]["alg"][-3:]
 
@@ -280,23 +283,25 @@ class AuthorizationResponse(message.AuthorizationResponse,
                 try:
                     assert "at_hash" in idt
                 except AssertionError:
-                    raise MissingRequiredAttribute("Missing at_hash property")
+                    raise MissingRequiredAttribute("Missing at_hash property",
+                                                   idt)
                 try:
                     assert idt["at_hash"] == jws.left_hash(
                         self["access_token"], hfunc)
                 except AssertionError:
                     raise VerificationError(
-                        "Failed to verify access_token hash")
+                        "Failed to verify access_token hash", idt)
 
             if "code" in self:
                 try:
                     assert "c_hash" in idt
                 except AssertionError:
-                    raise MissingRequiredAttribute("Missing c_hash property")
+                    raise MissingRequiredAttribute("Missing c_hash property",
+                                                   idt)
                 try:
                     assert idt["c_hash"] == jws.left_hash(self["code"], hfunc)
                 except AssertionError:
-                    raise VerificationError("Failed to verify code hash")
+                    raise VerificationError("Failed to verify code hash", idt)
 
             self["id_token"] = idt
 
@@ -334,7 +339,8 @@ class AuthorizationRequest(message.AuthorizationRequest):
             "registration": SINGLE_OPTIONAL_JSON,
             "request": SINGLE_OPTIONAL_STRING,
             "request_uri": SINGLE_OPTIONAL_STRING,
-            "session_state": SINGLE_OPTIONAL_STRING,
+            #"session_state": SINGLE_OPTIONAL_STRING,
+            "response_mode": SINGLE_OPTIONAL_STRING,
         }
     )
     c_allowed_values = message.AuthorizationRequest.c_allowed_values.copy()
@@ -377,29 +383,30 @@ class AuthorizationRequest(message.AuthorizationRequest):
                 self["id_token"] = idt
 
         if "response_type" not in self:
-            raise MissingRequiredAttribute("response_type missing")
+            raise MissingRequiredAttribute("response_type missing", self)
 
         _rt = self["response_type"]
         if "token" in _rt or "id_token" in _rt:
             try:
                 assert "nonce" in self
             except AssertionError:
-                raise MissingRequiredAttribute("Nonce missing")
+                raise MissingRequiredAttribute("Nonce missing", self)
 
         try:
             assert "openid" in self["scope"]
         except AssertionError:
-            raise MissingRequiredValue("openid in scope")
+            raise MissingRequiredValue("openid in scope", self)
 
         if "offline_access" in self["scope"]:
             try:
                 assert "consent" in self["prompt"]
             except AssertionError:
-                raise MissingRequiredValue("consent in prompt")
+                raise MissingRequiredValue("consent in prompt", self)
 
         if "prompt" in self:
             if "none" in self["prompt"] and len(self["prompt"]) > 1:
-                raise InvalidRequest("prompt none combined with other value")
+                raise InvalidRequest("prompt none combined with other value",
+                                     self)
 
         return super(AuthorizationRequest, self).verify(**kwargs)
 
@@ -457,7 +464,7 @@ class OpenIDSchema(Message):
                 try:
                     _ = time.strptime(self["birthdate"], "%Y")
                 except ValueError:
-                    raise VerificationError("Birthdate format error")
+                    raise VerificationError("Birthdate format error", self)
 
         return super(OpenIDSchema, self).verify(**kwargs)
 
@@ -474,6 +481,7 @@ class RegistrationRequest(Message):
         "client_uri": SINGLE_OPTIONAL_STRING,
         "policy_uri": SINGLE_OPTIONAL_STRING,
         "tos_uri": SINGLE_OPTIONAL_STRING,
+        "jwks": SINGLE_OPTIONAL_STRING,
         "jwks_uri": SINGLE_OPTIONAL_STRING,
         "sector_identifier_uri": SINGLE_OPTIONAL_STRING,
         "subject_type": SINGLE_OPTIONAL_STRING,
@@ -495,7 +503,7 @@ class RegistrationRequest(Message):
         "request_uris": OPTIONAL_LIST_OF_STRINGS,
         #"client_id": SINGLE_OPTIONAL_STRING,
         #"client_secret": SINGLE_OPTIONAL_STRING,
-        "access_token": SINGLE_OPTIONAL_STRING,
+        #"access_token": SINGLE_OPTIONAL_STRING,
         "post_logout_redirect_uris": OPTIONAL_LIST_OF_STRINGS,
     }
     c_default = {"application_type": "web"}
@@ -535,11 +543,11 @@ class RegistrationResponse(Message):
             if not "registration_access_token":
                 raise VerificationError((
                     "Only one of registration_client_uri"
-                    " and registration_access_token present"))
+                    " and registration_access_token present"), self)
         elif "registration_access_token" in self:
             raise VerificationError((
                 "Only one of registration_client_uri"
-                " and registration_access_token present"))
+                " and registration_access_token present"), self)
 
         return super(RegistrationResponse, self).verify(**kwargs)
 
@@ -572,7 +580,25 @@ class IdToken(OpenIDSchema):
             if "client_id" in kwargs:
                 # check that I'm among the recipients
                 if kwargs["client_id"] not in self["aud"]:
-                    return False
+                    raise NotForMe("", self)
+
+            if len(self["aud"]) > 1:  # Then azr has to be present and be one of
+                                      # the values
+                try:
+                    assert "azr" in self
+                except AssertionError:
+                    raise VerificationError("azr missing", self)
+                else:
+                    try:
+                        assert self["azr"] in self["aud"]
+                    except AssertionError:
+                        raise VerificationError(
+                            "Missmatch between azr and aud claims", self)
+
+        if "azr" in self:
+            if "client_id" in kwargs:
+                if kwargs["client_id"] != self["azp"]:
+                    raise NotForMe("", self)
 
         return super(IdToken, self).verify(**kwargs)
 
@@ -597,9 +623,10 @@ class CheckIDRequest(Message):
 
 
 class EndSessionRequest(Message):
-    c_param = {"id_token": SINGLE_REQUIRED_STRING,
-               "redirect_url": SINGLE_REQUIRED_STRING,
-               "state": SINGLE_REQUIRED_STRING}
+    c_param = {
+        "id_token_hint": SINGLE_OPTIONAL_STRING,
+        "post_logout_redirect_uri": SINGLE_OPTIONAL_STRING
+    }
 
 
 class EndSessionResponse(Message):
@@ -633,7 +660,6 @@ class OpenIDRequest(AuthorizationRequest):
 
 class ProviderConfigurationResponse(Message):
     c_param = {
-        "version": SINGLE_REQUIRED_STRING,
         "issuer": SINGLE_REQUIRED_STRING,
         "authorization_endpoint": SINGLE_OPTIONAL_STRING,
         "token_endpoint": SINGLE_OPTIONAL_STRING,
@@ -679,7 +705,7 @@ class ProviderConfigurationResponse(Message):
         #"x509_encryption_url": SINGLE_OPTIONAL_STRING,
     }
     c_default = {"version": "3.0",
-                 "token_endpoint_auth_methods_supported": "client_secret_basic",
+                 "token_endpoint_auth_methods_supported": ["client_secret_basic"],
                  "claims_parameter_supported": "false",
                  "request_parameter_supported": "false",
                  "request_uri_parameter_supported": "true",
@@ -693,7 +719,11 @@ class ProviderConfigurationResponse(Message):
                 check_char_set(scope, SCOPE_CHARSET)
 
         parts = urlparse(self["issuer"])
-        assert parts.scheme == "https"
+        try:
+            assert parts.scheme == "https"
+        except AssertionError:
+            raise SchemeError("Not HTTPS")
+
         assert not parts.query and not parts.fragment
 
         return super(ProviderConfigurationResponse, self).verify(**kwargs)
