@@ -1,9 +1,11 @@
 #!/usr/bin/env python
+import copy
 import json
 import traceback
 import urllib
 import sys
 from jwkest.jwe import JWE
+from oic.utils import time_util
 from oic.utils.authn.user import NoSuchAuthentication
 from oic.utils.authn.user import ToOld
 from oic.utils.authn.user import TamperAllert
@@ -14,9 +16,10 @@ from oic.utils.keyio import key_export
 from requests import ConnectionError
 
 from oic.oauth2.message import by_schema
-from oic.oauth2.message import MessageException
 from oic.oic.message import RefreshAccessTokenRequest
-from oic.oic.message import AuthorizationRequest, Claims
+from oic.oic.message import EndSessionRequest
+from oic.oic.message import AuthorizationRequest
+from oic.oic.message import Claims
 from oic.oic.message import IdToken
 from oic.oic.message import OpenIDSchema
 from oic.oic.message import RegistrationResponse
@@ -58,7 +61,8 @@ from oic.oauth2 import rndstr
 
 from oic.oic import Server
 
-from oic.oauth2.exception import *
+from oic.exception import *
+from jwkest.jwe import JWEException
 
 logger = logging.getLogger(__name__)
 
@@ -106,7 +110,7 @@ def code_token_response(**kwargs):
 
     aresp["code"] = _scode
 
-    _dic = _sdb.update_to_token(_scode, issue_refresh=False)
+    _dic = _sdb.upgrade_to_token(_scode, issue_refresh=False)
     for prop in AccessTokenResponse.c_param.keys():
         try:
             aresp[prop] = _dic[prop]
@@ -153,21 +157,25 @@ class EndSessionEndpoint(Endpoint) :
 class Provider(AProvider):
     def __init__(self, name, sdb, cdb, authn_broker, userinfo, authz,
                  client_authn, symkey, urlmap=None, ca_certs="", keyjar=None,
-                 hostname="", template_lookup=None, verify_login_template=None):
+                 hostname="", template_lookup=None, template=None,
+                 verify_ssl=True, capabilities=None):
 
         AProvider.__init__(self, name, sdb, cdb, authn_broker, authz,
-                           client_authn, symkey, urlmap)
+                           client_authn, symkey, urlmap, ca_bundle=ca_certs,
+                           verify_ssl=verify_ssl)
+
+        # Should be a OIC Server not an OAuth2 server
+        self.server = Server(ca_certs=ca_certs, verify_ssl=verify_ssl)
 
         self.endp.extend([UserinfoEndpoint, RegistrationEndpoint,
                           EndSessionEndpoint])
 
         self.userinfo = userinfo
-        self.server = Server(ca_certs=ca_certs)
 
         if keyjar:
             self.server.keyjar = keyjar
         self.template_lookup = template_lookup
-        self.verify_login_template = verify_login_template
+        self.template = template or {}
         self.keyjar = self.server.keyjar
         self.baseurl = ""
         self.cert = []
@@ -184,12 +192,18 @@ class Provider(AProvider):
         self.hostname = hostname or socket.gethostname
         self.register_endpoint = "%s%s" % (self.baseurl, "register")
 
+        if capabilities:
+            self.verify_capabilities(capabilities)
+            self.capabilities = capabilities
+        else:
+            self.capabilities = self.provider_features()
+
     def id_token_as_signed_jwt(self, session, loa="2", alg="RS256", code=None,
-                               access_token=None, user_info=None):
+                               access_token=None, user_info=None, auth_time=0):
 
         logger.debug("Signing alg: %s [%s]" % (alg, alg2keytype(alg)))
-        _idt = self.server.make_id_token(session, loa, self.name, alg, code,
-                                         access_token, user_info)
+        _idt = self.server.make_id_token(session, loa, self.baseurl, alg, code,
+                                         access_token, user_info, auth_time)
 
         logger.debug("id_token: %s" % _idt.to_dict())
         # My signing key if its RS*, can use client secret if HS*
@@ -228,7 +242,7 @@ class Provider(AProvider):
         Pick the sector id given a number of factors
         :param redirect_uri: The redirect_uri used
         :param client_info: Information provided by the client in the
-          client registration
+        client registration
         :return: A sector_id or None
         """
 
@@ -342,98 +356,75 @@ class Provider(AProvider):
     def verify_post_logout_redirect_uri(self, areq, cookie):
         try:
             redirect_uri = areq["post_logout_redirect_uri"]
-            authn = self.pick_auth(areq)
+            authn, acr = self.pick_auth(areq)
             uid = authn.authenticated_as(cookie)["uid"]
             client_info = self.cdb[self.sdb.getClient_id(uid)]
-            for tmpUri1 in redirect_uri:
-                for tmpUri2 in client_info["post_logout_redirect_uris"]:
-                    if str(tmpUri1) == str(tmpUri2[0]):
-                        return tmpUri1
+            if redirect_uri in client_info["post_logout_redirect_uris"]:
+                return redirect_uri
         except:
             pass
         return None
 
     def is_session_revoked(self, request="", cookie=None):
         areq = urlparse.parse_qs(request)
-        redirect_uri = self.verify_post_logout_redirect_uri(areq, cookie)
-        authn = self.pick_auth(areq)
+        authn, acr = self.pick_auth(areq)
         identity = authn.authenticated_as(cookie)
         return self.sdb.is_revoke_uid(identity["uid"])
 
-    def end_session_endpoint(self, request="", cookie=None, **kwargs):
-        areq = None
-        redirect_uri = None
+    def let_user_verify_logout(self, uid, esr, cookie, redirect_uri):
+        if cookie:
+            headers = [cookie]
+        else:
+            headers = []
+        mte = self.template_lookup.get_template(self.template["verify_logout"])
+        self.sdb.set_verified_logout(uid)
+        if redirect_uri is not None:
+            redirect = redirect_uri
+        else:
+            redirect = "/"
         try:
-            areq = urlparse.parse_qs(request)
-            redirect_uri = self.verify_post_logout_redirect_uri(areq, cookie)
-            authn = self.pick_auth(areq)
-            identity = authn.authenticated_as(cookie)
-            if "uid" not in identity:
-                return self._error_response("Not allowed!")
+            tmp_id_token_hint = esr["id_token_hint"]
         except:
+            tmp_id_token_hint = ""
+        argv = {
+            "id_token_hint": tmp_id_token_hint,
+            "post_logout_redirect_uri": esr["post_logout_redirect_uri"],
+            "key": self.sdb.get_verify_logout(uid),
+            "redirect": redirect,
+            "action": "/"+EndSessionEndpoint("").etype
+        }
+        return Response(mte.render(**argv), headers=[])
+
+    def end_session_endpoint(self, request="", cookie=None, **kwargs):
+        esr = EndSessionRequest().from_urlencoded(request)
+        redirect_uri = self.verify_post_logout_redirect_uri(esr, cookie)
+        if not redirect_uri:
             return self._error_response("Not allowed!")
 
-        verify = self.sdb.getVerifyLogout(identity["uid"])
-        if (verify is None or "key" not in areq or verify != areq["key"][0]) and \
-                (self.template_lookup is not None and self.verify_login_template is not None):
-            if cookie:
-                headers = [cookie]
-            else:
-                headers = []
-            mte = self.template_lookup.get_template(self.verify_login_template)
-            self.sdb.setVerifyLogout(identity["uid"])
-            if redirect_uri is not None:
-                redirect = redirect_uri
-            else:
-                redirect = "/"
-            try:
-                tmp_id_token_hint = areq["id_token_hint"][0]
-            except:
-                tmp_id_token_hint = ""
-            argv = {
-                "id_token_hint": tmp_id_token_hint,
-                "post_logout_redirect_uri": areq["post_logout_redirect_uri"][0],
-                "key": self.sdb.getVerifyLogout(identity["uid"]),
-                "redirect": redirect,
-                "action": "/"+EndSessionEndpoint("").etype
-            }
-            #resp.message = mte.render(**argv)
-            return Response(mte.render(**argv), headers=[])
+        authn, acr = self.pick_auth(esr)
 
-        id_token = None
-        try:
-            id_token = self.sdb.getToken_id(identity["uid"])
-        except:
-            pass
-
-        if id_token is not None and "id_token_hint" in areq:
+        if "id_token_hint" in esr:
+            id_token_hint = OpenIDRequest().from_jwt(esr["id_token_hint"],
+                                                     keyjar=self.keyjar,
+                                                         verify=True)
+            uid = id_token_hint["sub"]
+        else:
+            identity = authn.authenticated_as(cookie)
             try:
-                id_token_hint = OpenIDRequest().from_jwt(areq["id_token_hint"][0], keyjar=self.keyjar, verify=True)
-                id_token = OpenIDRequest().from_jwt(id_token, keyjar=self.keyjar, verify=True)
-                id_token_hint_dict = id_token_hint.to_dict()
-                id_token_dict = id_token.to_dict()
-                for key in id_token_dict:
-                    if key in id_token_hint_dict:
-                        if id_token_dict[key] != id_token_hint_dict[key]:
-                            return self._error_response("Not allowed!")
-                    else:
-                        return self._error_response("Not allowed!")
-                for key in id_token_hint_dict:
-                    if key in id_token_dict:
-                        if id_token_dict[key] != id_token_hint_dict[key]:
-                            return self._error_response("Not allowed!")
-                    else:
-                        return self._error_response("Not allowed!")
-            except:
-                self._error_response("Not allowed!")
-        elif id_token is not None:
-            self._error_response("Not allowed!")
+                uid = identity["uid"]
+            except KeyError:
+                return self._error_response("Not allowed!")
+
+        #if self.sdb.get_verified_logout(uid):
+        #    return self.let_user_verify_logout(uid, esr, cookie, redirect_uri)
 
         try:
-            self.sdb.revoke_uid(identity["uid"])
-        except:
+            sid = self.sdb.get_sid_from_userid(uid)
+        except KeyError:
             pass
             #If cleanup cannot be performed we will still invalidate the cookie.
+        else:
+            del self.sdb[sid]
 
         if redirect_uri is not None:
             return Redirect(str(redirect_uri), headers=[authn.delete_cookie()])
@@ -441,13 +432,32 @@ class Provider(AProvider):
         return Response("", headers=[authn.delete_cookie()])
 
     def verify_endpoint(self, request="", cookie=None, **kwargs):
+        """
+
+        :param request:
+        :param cookie:
+        :param kwargs:
+        :return:
+        """
+        logger.debug("verify request: %s" % request)
+
         _req = urlparse.parse_qs(request)
+        if "query" in _req:
+            try:
+                # TODO FIX THIS !!! Why query ?
+                areq = urlparse.parse_qs(_req["query"][0])
+            except KeyError:
+                return BadRequest()
+        else:
+            areq = _req
+
+        logger.debug("REQ: %s" % areq)
         try:
-            areq = urlparse.parse_qs(_req["query"][0])
-        except KeyError:
-            return BadRequest()
+            authn, acr = self.pick_auth(areq, "exact")
+        except Exception, err:
+            logger.exception("%s", err)
+            raise
 
-        authn, acr = self.pick_auth(areq, "exact")
         kwargs["cookie"] = cookie
         return authn.verify(_req, **kwargs)
 
@@ -514,6 +524,7 @@ class Provider(AProvider):
             identity = authn.authenticated_as(cookie,
                                               authorization=_auth_info,
                                               max_age=self.max_age(areq))
+            auth_time = time_util.utc_now()
         except (NoSuchAuthentication, ToOld, TamperAllert):
             identity = None
 
@@ -561,7 +572,8 @@ class Provider(AProvider):
         except KeyError:
             oidc_req = None
 
-        sid = self.sdb.create_authz_session(user, areq, oidreq=oidc_req)
+        sid = self.sdb.create_authz_session(user, areq, oidreq=oidc_req,
+                                            auth_time=auth_time)
         return self.authz_part2(user, areq, sid)
 
     def userinfo_in_id_token_claims(self, session):
@@ -598,8 +610,9 @@ class Provider(AProvider):
             enc = "A128CBC"
 
         keys = self.keyjar.get_encrypt_key(owner=cid)
-        logger.debug("keys for %s: %s" % (cid, self.keyjar[cid]))
-        logger.debug("alg=%s, enc=%s" % (alg, enc))
+        logger.debug("keys for %s: %s" % (
+            cid, "["+", ".join([str(x) for x in self.keyjar[cid]]))+"]")
+        logger.debug("alg=%s, enc=%s, val_type=%s" % (alg, enc, val_type))
         logger.debug("Encryption keys for %s: %s" % (cid, keys))
 
         # use the clients public key for encryption
@@ -607,7 +620,7 @@ class Provider(AProvider):
         return _jwe.encrypt(keys, context="public")
 
     def sign_encrypt_id_token(self, sinfo, client_info, areq, code=None,
-                              access_token=None, user_info=None):
+                              access_token=None, user_info=None, auth_time=0):
         """
         Signed and or encrypt a IDToken
 
@@ -628,7 +641,8 @@ class Provider(AProvider):
         id_token = self.id_token_as_signed_jwt(sinfo, alg=alg,
                                                code=code,
                                                access_token=access_token,
-                                               user_info=user_info)
+                                               user_info=user_info,
+                                               auth_time=auth_time)
 
         # Then encrypt
         if "id_token_encrypted_response_alg" in client_info:
@@ -665,7 +679,7 @@ class Provider(AProvider):
         _log_debug("All checks OK")
 
         try:
-            _tinfo = _sdb.update_to_token(_access_code)
+            _tinfo = _sdb.upgrade_to_token(_access_code)
         except Exception, err:
             logger.error("%s" % err)
             # Should revoke the token issued to this access code
@@ -675,7 +689,8 @@ class Provider(AProvider):
         if "openid" in _info["scope"]:
             userinfo = self.userinfo_in_id_token_claims(_info)
             _idtoken = self.sign_encrypt_id_token(_info, client_info, req,
-                                                  user_info=userinfo)
+                                                  user_info=userinfo,
+                                                  auth_time=_info["auth_time"])
             _sdb.update_by_token(_access_code, "id_token", _idtoken)
 
         _log_debug("_tinfo: %s" % _tinfo)
@@ -699,7 +714,8 @@ class Provider(AProvider):
         if "openid" in _info["scope"]:
             userinfo = self.userinfo_in_id_token_claims(_info)
             _idtoken = self.sign_encrypt_id_token(_info, client_info, req,
-                                                  user_info=userinfo)
+                                                  user_info=userinfo,
+                                                  auth_time=_info["auth_time"])
             sid = _sdb.token.get_key(rtoken)
             _sdb.update(sid, "id_token", _idtoken)
 
@@ -743,7 +759,12 @@ class Provider(AProvider):
             req["client_id"] = client_id
 
         if isinstance(req, AccessTokenRequest):
-            return self._access_token_endpoint(req, **kwargs)
+            try:
+                return self._access_token_endpoint(req, **kwargs)
+            except JWEException as err:
+                return self._error_response("invalid_request",
+                                            descr="%s" % err)
+
         else:
             return self._refresh_access_token_endpoint(req, **kwargs)
 
@@ -843,23 +864,30 @@ class Provider(AProvider):
 
         # Should I return a JSON or a JWT ?
         _cinfo = self.cdb[session["client_id"]]
-        if "userinfo_signed_response_alg" in _cinfo:
-            algo = _cinfo["userinfo_signed_response_alg"]
-            # Use my key for signing
-            key = self.keyjar.get_signing_key(alg2keytype(algo), "")
-            jinfo = info.to_jwt(key, algo)
-            content_type = "application/jwt"
-            if "userinfo_encrypted_response_alg" in _cinfo:
-                # encrypt with clients public key
-                jinfo = self.encrypt(jinfo, _cinfo, session["client_id"],
+        try:
+            if "userinfo_signed_response_alg" in _cinfo:
+                algo = _cinfo["userinfo_signed_response_alg"]
+                # Use my key for signing
+                key = self.keyjar.get_signing_key(alg2keytype(algo), "")
+                if not key:
+                    return self._error(error="access_denied",
+                                       descr="Missing signing key")
+                jinfo = info.to_jwt(key, algo)
+                content_type = "application/jwt"
+                if "userinfo_encrypted_response_alg" in _cinfo:
+                    # encrypt with clients public key
+                    jinfo = self.encrypt(jinfo, _cinfo, session["client_id"],
+                                         "userinfo")
+            elif "userinfo_encrypted_response_alg" in _cinfo:
+                jinfo = self.encrypt(info.to_json(), _cinfo, session["client_id"],
                                      "userinfo")
-        elif "userinfo_encrypted_response_alg" in _cinfo:
-            jinfo = self.encrypt(info.to_json(), _cinfo, session["client_id"],
-                                 "userinfo")
-            content_type = "application/jwt"
-        else:
-            jinfo = info.to_json()
-            content_type = "application/json"
+                content_type = "application/jwt"
+            else:
+                jinfo = info.to_json()
+                content_type = "application/json"
+        except JWEException:
+            return self._error(error="access_denied",
+                               descr="Could not encrypt")
 
         return Response(jinfo, content=content_type)
 
@@ -1019,6 +1047,7 @@ class Provider(AProvider):
                 pass
         except Exception, err:
             logger.error("Failed to load client keys: %s" % request.to_dict())
+            logger.error("%s", err)
             err = ClientRegistrationErrorResponse(
                 error="invalid_configuration_parameter",
                 error_description="%s" % err)
@@ -1060,9 +1089,12 @@ class Provider(AProvider):
         _log_debug = logger.debug
         _log_info = logger.info
 
-        _log_debug("@registration_endpoint")
+        _log_debug("@registration_endpoint: <<%s>>" % request)
 
-        request = RegistrationRequest().deserialize(request, "json")
+        try:
+            request = RegistrationRequest().deserialize(request, "json")
+        except ValueError:
+            request = RegistrationRequest().deserialize(request)
 
         _log_info("registration_request:%s" % request.to_dict())
         resp_keys = request.keys()
@@ -1117,9 +1149,7 @@ class Provider(AProvider):
         self.comb_post_logout_redirect_uris(args)
         response = RegistrationResponse(**args)
 
-        self.keyjar.load_keys(request, client_id)
-
-        # Add the key to the keyjar
+        # Add the client_secret as a symmetric key to the keyjar
         if client_secret:
             _kc = KeyBundle([{"kty": "oct", "key": client_secret,
                               "use": "ver"},
@@ -1176,9 +1206,56 @@ class Provider(AProvider):
         return Response(response.to_json(), content="application/json",
                         headers=[("Cache-Control", "no-store")])
 
-    def create_providerinfo(self, pcr_class=ProviderConfigurationResponse):
-        _response = pcr_class(
-            issuer=self.baseurl,
+    def create_providerinfo(self, pcr_class=ProviderConfigurationResponse,
+                            setup=None):
+        """
+        Dynamically create the provider info response
+        :param pcr_class:
+        :param setup:
+        :return:
+        """
+
+        _provider_info = copy.deepcopy(self.capabilities)
+
+        if setup:
+            # sort of placeholder tight now
+            # for key, spec in pcr_class.c_param.items():
+            #     if isinstance(setup[0], list) and setup[0] != basestring:
+            #         # dealing with lists
+            #         if key in setup:
+            #             if key in _provider_info:
+            #                 _provider_info[key] = [x for x in setup[key]
+            #                                        if x in _provider_info[key]]
+            #             else:
+            #                 _provider_info[key] = setup[key]
+            #     else:
+            pass
+        else:
+            #keys = self.keyjar.keys_by_owner(owner=".")
+            if self.jwks_uri and self.keyjar:
+                _provider_info["jwks_uri"] = self.jwks_uri
+
+            for endp in self.endp:
+                #_log_info("# %s, %s" % (endp, endp.name))
+                _provider_info[endp(None).name] = "%s%s" % (self.baseurl,
+                                                            endp.etype)
+
+        _provider_info["issuer"] = self.baseurl
+        _provider_info["version"] = "3.0"
+
+        if not self.baseurl.endswith("/"):
+            self.baseurl += "/"
+
+        return _provider_info
+
+    def provider_features(self, pcr_class=ProviderConfigurationResponse):
+        """
+        Specifies what the server capabilities are.
+
+        :param pcr_class:
+        :return: ProviderConfigurationResponse instance
+        """
+        _provider_info = pcr_class(
             token_endpoint_auth_methods_supported=[
                 "client_secret_post", "client_secret_basic",
                 "client_secret_jwt", "private_key_jwt"],
@@ -1187,6 +1264,7 @@ class Provider(AProvider):
                                       "code token", "code id_token",
                                       "token id_token",
                                       "code token id_token"],
+            response_modes_supported=['query', 'fragment', 'form_post'],
             subject_types_supported=["public", "pairwise"],
             grant_types_supported=[
                 "authorization_code", "implicit",
@@ -1199,37 +1277,48 @@ class Provider(AProvider):
         )
 
         sign_algs = jws.SIGNER_ALGS.keys()
-
         for typ in ["userinfo", "id_token", "request_object",
                     "token_endpoint_auth"]:
-            _response["%s_signing_alg_values_supported" % typ] = sign_algs
+            _provider_info["%s_signing_alg_values_supported" % typ] = sign_algs
 
         algs = jwe.SUPPORTED["alg"]
         for typ in ["userinfo", "id_token", "request_object"]:
-            _response["%s_encryption_alg_values_supported" % typ] = algs
+            _provider_info["%s_encryption_alg_values_supported" % typ] = algs
 
         encs = jwe.SUPPORTED["enc"]
         for typ in ["userinfo", "id_token", "request_object"]:
-            _response["%s_encryption_enc_values_supported" % typ] = encs
-
-        if not self.baseurl.endswith("/"):
-            self.baseurl += "/"
-
-        #keys = self.keyjar.keys_by_owner(owner=".")
-        if self.jwks_uri and self.keyjar:
-            _response["jwks_uri"] = self.jwks_uri
+            _provider_info["%s_encryption_enc_values_supported" % typ] = encs
 
         #acr_values
         if self.authn_broker:
             acr_values = self.authn_broker.getAcrValuesString()
             if acr_values is not None:
-                _response["acr_values_supported"] = acr_values
+                _provider_info["acr_values_supported"] = acr_values
 
-        for endp in self.endp:
-            #_log_info("# %s, %s" % (endp, endp.name))
-            _response[endp(None).name] = "%s%s" % (self.baseurl, endp.etype)
+        return _provider_info
 
-        return _response
+    def verify_capabilities(self, capabilities):
+        """
+        Verify that what the admin wants the server to do actually
+        can be done by this implementation.
+
+        :param capabilities: The asked for capabilities as a dictionary
+        or a ProviderConfigurationResponse instance. The later can be
+        treated as a dictionary.
+        :return: True or False
+        """
+        _pinfo = self.provider_features()
+        for key, val in capabilities:
+            if isinstance(val, basestring):
+                try:
+                    if val in _pinfo[key]:
+                        continue
+                    else:
+                        return False
+                except KeyError:
+                    return False
+
+        return True
 
     #noinspection PyUnusedLocal
     def providerinfo_endpoint(self, handle="", **kwargs):
@@ -1320,7 +1409,7 @@ class Provider(AProvider):
         if "response_type" in areq and \
                 len(areq["response_type"]) == 1 and \
                 "none" in areq["response_type"]:
-            pass
+            fragment_enc = False
         else:
             if self.sdb.is_revoked(sid):
                 return self._error(error="access_denied",
@@ -1336,6 +1425,11 @@ class Provider(AProvider):
             _log_debug("_dic: %s" % _sinfo)
 
             rtype = set(areq["response_type"][:])
+            if len(rtype) == 1 and "code" in rtype:
+                fragment_enc = False
+            else:
+                fragment_enc = True
+
             if "code" in areq["response_type"]:
                 #if issue_new_code:
                 #    scode = self.sdb.duplicate(_sinfo)
@@ -1348,7 +1442,7 @@ class Provider(AProvider):
                 _code = None
 
             if "token" in rtype:
-                _dic = self.sdb.update_to_token(issue_refresh=False, key=sid)
+                _dic = self.sdb.upgrade_to_token(issue_refresh=False, key=sid)
 
                 _log_debug("_dic: %s" % _dic)
                 for key, val in _dic.items():
@@ -1368,7 +1462,8 @@ class Provider(AProvider):
 
                 id_token = self.sign_encrypt_id_token(
                     _sinfo, client_info, areq, code=_code,
-                    access_token=_access_token, user_info=user_info)
+                    access_token=_access_token, user_info=user_info,
+                    auth_time=_sinfo["auth_time"])
 
                 aresp["id_token"] = id_token
                 _sinfo["id_token"] = id_token
@@ -1394,7 +1489,28 @@ class Provider(AProvider):
 
         # so everything went well should set a SSO cookie
         headers = [self.cookie_func(user, typ="sso", ttl=self.sso_ttl)]
-        location = aresp.request(redirect_uri)
+
+        # Now about the response_mode. Should not be set if it's obvious
+        # from the response_type. Knows about 'query', 'fragment' and
+        # 'form_post'.
+
+        if "response_mode" in areq:
+            resp_mode = areq["response_mode"]
+            if resp_mode == "form_post":
+                argv = {"form_args": aresp.to_dict(),
+                        "action": redirect_uri}
+                mte = self.template_lookup.get_template(
+                    self.template["form_post"])
+                return Response(mte.render(**argv), headers=headers)
+            elif resp_mode == 'fragment' and not fragment_enc:
+                # Can't be done
+                return self._error("invalid_request", "wrong response_mode")
+            elif resp_mode == 'query' and fragment_enc:
+                # Can't be done
+                return self._error("invalid_request", "wrong response_mode")
+
+        # Just do whatever is the default
+        location = aresp.request(redirect_uri, fragment_enc)
         logger.debug("Redirected to: '%s' (%s)" % (location, type(location)))
         return Redirect(str(location), headers=headers)
 
@@ -1413,6 +1529,15 @@ class Provider(AProvider):
     def register_endpoint(self, request="", **kwargs):
         pass
 
+    def endsession_endpoint(self, request="", **kwargs):
+        """
+
+        :param request:
+        :param kwargs:
+        :return: Either a Response instance or a tuple (Response, args)
+        """
+        return self.end_session_endpoint(request, **kwargs)
+
 # -----------------------------------------------------------------------------
 
 
