@@ -1,6 +1,7 @@
 import json
 import time
 from Crypto.PublicKey import RSA
+from oic.exception import MessageException
 
 __author__ = 'rohe0002'
 
@@ -41,12 +42,12 @@ class KeyBundle(object):
                  fileformat="jwk", keytype="RSA", keyusage=None):
         """
 
-        :param keys: A list of dictionaries of the format
+        :param keys: A list of dictionaries
             with the keys ["kty", "key", "alg", "use", "kid"]
         :param source: Where the key set can be fetch from
         :param verify_ssl: Verify the SSL cert used by the server
         :param fileformat: For a local file either "jwk" or "der"
-        :param keytype: Iff local file and 'der' format what kind of key is it.
+        :param keytype: Iff local file and 'der' format what kind of key it is.
         """
 
         self._keys = []
@@ -101,6 +102,8 @@ class KeyBundle(object):
                     continue
                 else:
                     _key.dc()
+                    if _typ == "EC":
+                        _key.ser = True
                     self._keys.append(_key)
                     flag = 1
                     break
@@ -121,7 +124,10 @@ class KeyBundle(object):
         for use in keyusage:
             _key = K2C[keytype]()
             _key.key = _bkey
-            _key.deserialize()
+
+            if _bkey:
+                _key.serialize()
+
             _key.use = use
             self._keys.append(_key)
 
@@ -165,7 +171,6 @@ class KeyBundle(object):
         """
         if self.source:
             # reread everything
-
             self._keys = []
 
             if self.remote is False:
@@ -225,6 +230,14 @@ class KeyBundle(object):
         for key in self._keys:
             if key.kid == kid:
                 return key
+
+        # Try updating since there might have been an update to the key file
+        self.update()
+
+        for key in self._keys:
+            if key.kid == kid:
+                return key
+
         return None
 
 
@@ -257,7 +270,8 @@ def dump_jwks(kbl, target):
     """
     res = {"keys": []}
     for kb in kbl:
-        res["keys"].extend([k.to_dict() for k in kb.keys()])
+        # ignore simple keys
+        res["keys"].extend([k.to_dict() for k in kb.keys() if k.kty != 'oct'])
 
     try:
         f = open(target, 'w')
@@ -274,10 +288,17 @@ def dump_jwks(kbl, target):
 class KeyJar(object):
     """ A keyjar contains a number of KeyBundles """
 
-    def __init__(self, ca_certs=None):
+    def __init__(self, ca_certs=None, verify_ssl=True):
+        """
+
+        :param ca_certs:
+        :param verify_ssl: Attempting SSL certificate verification
+        :return:
+        """
         self.spec2key = {}
         self.issuer_keys = {}
         self.ca_certs = ca_certs
+        self.verify_ssl = verify_ssl
 
     def add_if_unique(self, issuer, use, keys):
         if use in self.issuer_keys[issuer] and self.issuer_keys[issuer][use]:
@@ -302,7 +323,7 @@ class KeyJar(object):
         if "/localhost:" in url or "/localhost/" in url:
             kc = KeyBundle(source=url, verify_ssl=False)
         else:
-            kc = KeyBundle(source=url)
+            kc = KeyBundle(source=url, verify_ssl=self.verify_ssl)
 
         try:
             self.issuer_keys[issuer].append(kc)
@@ -334,12 +355,13 @@ class KeyJar(object):
 
         self.issuer_keys[issuer] = val
 
-    def get(self, use, key_type="", issuer=""):
+    def get(self, use, key_type="", issuer="", kid=None):
         """
 
         :param use: A key useful for this usage (enc, dec, sig, ver)
         :param key_type: Type of key (rsa, ec, symmetric, ..)
         :param issuer: Who is responsible for the keys, "" == me
+        :param kid: A Key Identifier
         :return: A possibly empty list of keys
         """
 
@@ -373,21 +395,26 @@ class KeyJar(object):
                 else:
                     _keys = bundles.keys()
                 for key in _keys:
-                    if use == key.use:
+                    if kid and key.kid == kid:
+                        lst = [key]
+                        break
+                    if not key.use or use == key.use:
                         lst.append(key)
+                if kid and lst:
+                    break
         return lst
 
-    def get_signing_key(self, key_type="", owner=""):
-        return self.get("sig", key_type, owner)
+    def get_signing_key(self, key_type="", owner="", kid=None):
+        return self.get("sig", key_type, owner, kid)
 
-    def get_verify_key(self, key_type="", owner=""):
-        return self.get("ver", key_type, owner)
+    def get_verify_key(self, key_type="", owner="", kid=None):
+        return self.get("ver", key_type, owner, kid)
 
-    def get_encrypt_key(self, key_type="", owner=""):
-        return self.get("enc", key_type, owner)
+    def get_encrypt_key(self, key_type="", owner="", kid=None):
+        return self.get("enc", key_type, owner, kid)
 
-    def get_decrypt_key(self, key_type="", owner=""):
-        return self.get("dec", key_type, owner)
+    def get_decrypt_key(self, key_type="", owner="", kid=None):
+        return self.get("dec", key_type, owner, kid)
 
     def get_key_by_kid(self, kid, owner=""):
         """
@@ -489,7 +516,11 @@ class KeyJar(object):
         """
 
         logger.debug("loading keys for issuer: %s" % issuer)
-        logger.debug("pcr: %s" % pcr)
+        try:
+            logger.debug("pcr: %s" % pcr)
+        except MessageException:
+            pass
+
         if issuer not in self.issuer_keys:
             self.issuer_keys[issuer] = []
         elif replace:
@@ -498,7 +529,12 @@ class KeyJar(object):
         try:
             self.add(issuer, pcr["jwks_uri"])
         except KeyError:
-            pass
+            #  jwks should only be considered if no jwks_uri is present
+            try:
+                _keys = pcr["jwks"]["keys"]
+                self.issuer_keys[issuer].append(KeyBundle(_keys))
+            except KeyError:
+                pass
 
     def find(self, source, issuer):
         """
@@ -513,10 +549,29 @@ class KeyJar(object):
         except KeyError:
             return None
 
+    def dump_issuer_keys(self, issuer):
+        res = []
+        try:
+            for kb in self.issuer_keys[issuer]:
+                res.extend([k.to_dict() for k in kb.keys()])
+        except KeyError:
+            pass
 
+        return res
+
+    def dump(self):
+        res = {}
+        for issuer in self.issuer_keys.keys():
+            res[issuer] = self.dump_issuer_keys(issuer)
+        return res
+
+    def restore(self, info):
+        for issuer, keys in info.items():
+            self.issuer_keys[issuer] = [KeyBundle(keys)]
 
 # =============================================================================
 
+
 class RedirectStdStreams(object):
     def __init__(self, stdout=None, stderr=None):
         self._stdout = stdout or sys.stdout
